# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import io
import struct

from . import multiplex as mp

class Register(enum.IntEnum):
    """These are the registers which are exposed for reading or writing
    from the moteus controller.

    The full list can be found at:
    https://github.com/mjbots/moteus/blob/main/docs/reference.md#a2b-registers

    """

    MODE = 0x000
    POSITION = 0x001
    VELOCITY = 0x002
    TORQUE = 0x003
    Q_CURRENT = 0x004
    D_CURRENT = 0x005
    ABS_POSITION = 0x006
    POWER = 0x007
    MOTOR_TEMPERATURE = 0x00a
    TRAJECTORY_COMPLETE = 0x00b
    REZERO_STATE = 0x00c
    HOME_STATE = 0x00c
    VOLTAGE = 0x00d
    TEMPERATURE = 0x00e
    FAULT = 0x00f

    PWM_PHASE_A = 0x010
    PWM_PHASE_B = 0x011
    PWM_PHASE_C = 0x012

    VOLTAGE_PHASE_A = 0x014
    VOLTAGE_PHASE_B = 0x015
    VOLTAGE_PHASE_C = 0x016

    VFOC_THETA = 0x018
    VFOC_VOLTAGE = 0x019
    VOLTAGEDQ_D = 0x01a
    VOLTAGEDQ_Q = 0x01b

    COMMAND_Q_CURRENT = 0x01c
    COMMAND_D_CURRENT = 0x01d

    VFOC_THETA_RATE = 0x01e

    COMMAND_POSITION = 0x020
    COMMAND_VELOCITY = 0x021
    COMMAND_FEEDFORWARD_TORQUE = 0x022
    COMMAND_KP_SCALE = 0x023
    COMMAND_KD_SCALE = 0x024
    COMMAND_POSITION_MAX_TORQUE = 0x025
    COMMAND_STOP_POSITION = 0x026
    COMMAND_TIMEOUT = 0x027
    COMMAND_VELOCITY_LIMIT = 0x028
    COMMAND_ACCEL_LIMIT = 0x029
    COMMAND_FIXED_VOLTAGE_OVERRIDE = 0x02a
    COMMAND_ILIMIT_SCALE = 0x02b
    COMMAND_FIXED_CURRENT_OVERRIDE = 0x02c
    COMMAND_IGNORE_POSITION_BOUNDS = 0x02d

    POSITION_KP = 0x030
    POSITION_KI = 0x031
    POSITION_KD = 0x032
    POSITION_FEEDFORWARD = 0x033
    POSITION_COMMAND = 0x034

    CONTROL_POSITION = 0x038
    CONTROL_VELOCITY = 0x039
    CONTROL_TORQUE = 0x03a
    POSITION_ERROR = 0x03b
    VELOCITY_ERROR = 0x03c
    TORQUE_ERROR = 0x03d

    COMMAND_WITHIN_LOWER_BOUND = 0x040
    COMMAND_WITHIN_UPPER_BOUND = 0x041
    COMMAND_WITHIN_FEEDFORWARD_TORQUE = 0x042
    COMMAND_WITHIN_KP_SCALE = 0x043
    COMMAND_WITHIN_KD_SCALE = 0x044
    COMMAND_WITHIN_MAX_TORQUE = 0x045
    COMMAND_WITHIN_TIMEOUT = 0x046
    COMMAND_WITHIN_ILIMIT_SCALE = 0x047
    COMMAND_WITHIN_IGNORE_POSITION_BOUNDS = 0x048

    ENCODER_0_POSITION = 0x050
    ENCODER_0_VELOCITY = 0x051
    ENCODER_1_POSITION = 0x052
    ENCODER_1_VELOCITY = 0x053
    ENCODER_2_POSITION = 0x054
    ENCODER_2_VELOCITY = 0x055

    ENCODER_VALIDITY = 0x058
    AUX1_GPIO_COMMAND = 0x05c
    AUX2_GPIO_COMMAND = 0x05d
    AUX1_GPIO_STATUS = 0x05e
    AUX2_GPIO_STATUS = 0x05f

    AUX1_ANALOG_IN1 = 0x060
    AUX1_ANALOG_IN2 = 0x061
    AUX1_ANALOG_IN3 = 0x062
    AUX1_ANALOG_IN4 = 0x063
    AUX1_ANALOG_IN5 = 0x064

    AUX2_ANALOG_IN1 = 0x068
    AUX2_ANALOG_IN2 = 0x069
    AUX2_ANALOG_IN3 = 0x06a
    AUX2_ANALOG_IN4 = 0x06b
    AUX2_ANALOG_IN5 = 0x06c

    MILLISECOND_COUNTER = 0x070
    CLOCK_TRIM = 0x071

    AUX1_PWM1 = 0x076,
    AUX1_PWM2 = 0x077,
    AUX1_PWM3 = 0x078,
    AUX1_PWM4 = 0x079,
    AUX1_PWM5 = 0x07a,
    AUX2_PWM1 = 0x07b,
    AUX2_PWM2 = 0x07c,
    AUX2_PWM3 = 0x07d,
    AUX2_PWM4 = 0x07e,
    AUX2_PWM5 = 0x07f,

    MODEL_NUMBER = 0x0100
    FIRMWARE_VERSION = 0x101
    REGISTER_MAP_VERSION = 0x102
    MULTIPLEX_ID = 0x110

    SERIAL_NUMBER = 0x120
    SERIAL_NUMBER1 = 0x120
    SERIAL_NUMBER2 = 0x121
    SERIAL_NUMBER3 = 0x122

    REZERO = 0x130
    SET_OUTPUT_NEAREST = 0x130
    SET_OUTPUT_EXACT = 0x131
    REQUIRE_REINDEX = 0x132
    RECAPTURE_POSITION_VELOCITY = 0x133

    DRIVER_FAULT1 = 0x140
    DRIVER_FAULT2 = 0x141

    UUID1 = 0x150
    UUID2 = 0x151
    UUID3 = 0x152
    UUID4 = 0x153

    UUID_MASK1 = 0x0154
    UUID_MASK2 = 0x0155
    UUID_MASK3 = 0x0156
    UUID_MASK4 = 0x0157

    UUID_MASK_CAPABLE = 0x0158


class Mode(enum.IntEnum):
    """Valid values for the Register.MODE register"""

    STOPPED = 0
    FAULT = 1
    PWM = 5
    VOLTAGE = 6
    VOLTAGE_FOC = 7
    VOLTAGE_DQ = 8
    CURRENT = 9
    POSITION = 10
    TIMEOUT = 11
    ZERO_VELOCITY = 12
    STAY_WITHIN = 13
    MEASURE_IND = 14
    BRAKE = 15


class Parser(mp.RegisterParser):
    def read_position(self, resolution):
        return self.read_mapped(resolution, 0.01, 0.0001, 0.00001)

    def read_velocity(self, resolution):
        return self.read_mapped(resolution, 0.1, 0.00025, 0.00001)

    def read_accel(self, resolution):
        return self.read_mapped(resolution, 0.05, 0.001, 0.00001)

    def read_torque(self, resolution):
        return self.read_mapped(resolution, 0.5, 0.01, 0.001)

    def read_pwm(self, resolution):
        return self.read_mapped(
            resolution, 1.0 / 127.0, 1.0 / 32767.0, 1.0 / 2147483647.0)

    def read_voltage(self, resolution):
        return self.read_mapped(resolution, 0.5, 0.1, 0.001)

    def read_temperature(self, resolution):
        return self.read_mapped(resolution, 1.0, 0.1, 0.001)

    def read_time(self, resolution):
        return self.read_mapped(resolution, 0.01, 0.001, 0.000001)

    def read_current(self, resolution):
        return self.read_mapped(resolution, 1.0, 0.1, 0.001)

    def read_power(self, resolution):
        return self.read_mapped(resolution, 10.0, 0.05, 0.0001)

    def ignore(self, resolution):
        self._offset += mp.resolution_size(resolution)


class Writer(mp.WriteFrame):
    def write_position(self, value, resolution):
        self.write_mapped(value, 0.01, 0.0001, 0.00001, resolution)

    def write_velocity(self, value, resolution):
        self.write_mapped(value, 0.1, 0.00025, 0.00001, resolution)

    def write_accel(self, value, resolution):
        self.write_mapped(value, 0.05, 0.001, 0.00001, resolution)

    def write_torque(self, value, resolution):
        self.write_mapped(value, 0.5, 0.01, 0.001, resolution)

    def write_pwm(self, value, resolution):
        self.write_mapped(value,
                          1.0 / 127.0,
                          1.0 / 32767.0,
                          1.0 / 2147483647.0,
                          resolution)

    def write_voltage(self, value, resolution):
        self.write_mapped(value, 0.5, 0.1, 0.001, resolution)

    def write_temperature(self, value, resolution):
        self.write_mapped(value, 1.0, 0.1, 0.001, resolution)

    def write_time(self, value, resolution):
        self.write_mapped(value, 0.01, 0.001, 0.000001, resolution)

    def write_current(self, value, resolution):
        self.write_mapped(value, 1.0, 0.1, 0.001, resolution)

    def write_power(self, value, resolution):
        self.write_mapped(value, 10.0, 0.05, 0.0001, resolution)

    def write_int(self, value, resolution):
        self.write_mapped(int(value), 1, 1, 1, resolution)


def parse_register(parser, register, resolution):
    if register == Register.MODE:
        return parser.read_int(resolution)
    elif register == Register.POSITION:
        return parser.read_position(resolution)
    elif register == Register.VELOCITY:
        return parser.read_velocity(resolution)
    elif register == Register.TORQUE:
        return parser.read_torque(resolution)
    elif register == Register.Q_CURRENT:
        return parser.read_current(resolution)
    elif register == Register.D_CURRENT:
        return parser.read_current(resolution)
    elif register == Register.ABS_POSITION:
        return parser.read_position(resolution)
    elif register == Register.POWER:
        return parser.read_power(resolution)
    elif register == Register.TRAJECTORY_COMPLETE:
        return parser.read_int(resolution)
    elif register == Register.HOME_STATE or register == Register.REZERO_STATE:
        return parser.read_int(resolution)
    elif register == Register.VOLTAGE:
        return parser.read_voltage(resolution)
    elif register == Register.MOTOR_TEMPERATURE:
        return parser.read_temperature(resolution)
    elif register == Register.TEMPERATURE:
        return parser.read_temperature(resolution)
    elif register == Register.FAULT:
        return parser.read_int(resolution)
    elif register == Register.POSITION_KP:
        return parser.read_torque(resolution)
    elif register == Register.POSITION_KI:
        return parser.read_torque(resolution)
    elif register == Register.POSITION_KD:
        return parser.read_torque(resolution)
    elif register == Register.POSITION_FEEDFORWARD:
        return parser.read_torque(resolution)
    elif register == Register.POSITION_COMMAND:
        return parser.read_torque(resolution)
    elif register == Register.CONTROL_POSITION:
        return parser.read_position(resolution)
    elif register == Register.CONTROL_VELOCITY:
        return parser.read_velocity(resolution)
    elif register == Register.CONTROL_TORQUE:
        return parser.read_torque(resolution)
    elif register == Register.POSITION_ERROR:
        return parser.read_position(resolution)
    elif register == Register.VELOCITY_ERROR:
        return parser.read_velocity(resolution)
    elif register == Register.TORQUE_ERROR:
        return parser.read_torque(resolution)
    elif register == Register.ENCODER_0_POSITION:
        return parser.read_position(resolution)
    elif register == Register.ENCODER_0_VELOCITY:
        return parser.read_velocity(resolution)
    elif register == Register.ENCODER_1_POSITION:
        return parser.read_position(resolution)
    elif register == Register.ENCODER_1_VELOCITY:
        return parser.read_velocity(resolution)
    elif register == Register.ENCODER_2_POSITION:
        return parser.read_position(resolution)
    elif register == Register.ENCODER_2_VELOCITY:
        return parser.read_velocity(resolution)
    elif register == Register.ENCODER_VALIDITY:
        return parser.read_int(resolution)
    elif register == Register.AUX1_GPIO_COMMAND:
        return parser.read_int(resolution)
    elif register == Register.AUX2_GPIO_COMMAND:
        return parser.read_int(resolution)
    elif register == Register.AUX1_GPIO_STATUS:
        return parser.read_int(resolution)
    elif register == Register.AUX2_GPIO_STATUS:
        return parser.read_int(resolution)
    elif (register == Register.AUX1_ANALOG_IN1 or
          register == Register.AUX1_ANALOG_IN2 or
          register == Register.AUX1_ANALOG_IN3 or
          register == Register.AUX1_ANALOG_IN4 or
          register == Register.AUX1_ANALOG_IN5 or
          register == Register.AUX2_ANALOG_IN1 or
          register == Register.AUX2_ANALOG_IN2 or
          register == Register.AUX2_ANALOG_IN3 or
          register == Register.AUX2_ANALOG_IN4 or
          register == Register.AUX2_ANALOG_IN5):
        return parser.read_pwm(resolution)
    elif register == Register.MILLISECOND_COUNTER:
        return parser.read_int(resolution)
    elif register == Register.CLOCK_TRIM:
        return parser.read_int(resolution)
    elif (register >= Register.AUX1_PWM1 and
          register <= Register.AUX2_PWM5):
        return parser.read_pwm(resolution)
    else:
        # We don't know what kind of value this is, so we don't know
        # the units.
        return parser.read(resolution)


def parse_reply(data):
    parser = Parser(data)
    result = {}
    while True:
        item = parser.next()
        if not item[0]:
            break
        resolution = item[2]
        register = item[1]
        result[register] = parse_register(parser, register, resolution)
    return result


class Result:
    id = None
    arbitration_id = None
    bus = None
    values = {}

    def __repr__(self):
        value_str = ', '.join(['{}(0x{:03x}): {}'.format(Register(key).name, key, value)
                              for key, value in self.values.items()])
        return f'{self.id}/{{{value_str}}}'

def parse_message(message):
    result = Result()
    result.id = (message.arbitration_id >> 8) & 0x7f
    result.values = parse_reply(message.data)

    # We store these things just for reference, so that our
    # results look a bit like CAN responses too.
    result.arbitration_id = message.arbitration_id
    if hasattr(message, 'bus'):
        result.bus = message.bus
    else:
        result.bus = 1
    result.data = message.data

    return result

def make_uuid_prefix(id):
    if isinstance(id, int) or id.uuid is None:
        return b''

    buf = io.BytesIO()
    writer = Writer(buf)

    reg_count = len(id.uuid) // 4
    assert len(id.uuid) % 4 == 0
    assert len(id.uuid) <= 16 and len(id.uuid) > 0

    combiner = mp.WriteCombiner(
        writer, 0x00, Register.UUID_MASK1, [mp.INT32] * reg_count)

    # Convert UUID prefix bytes to list of 32-bit integers
    # (little-endian)
    i32_data = []
    for i in range(0, len(id.uuid), 4):
        chunk = id.uuid[i:i+4]
        i32_val = struct.unpack('<i', chunk)[0]
        i32_data.append(i32_val)

    for i32_val in i32_data:
        if combiner.maybe_write():
            writer.write_int32(i32_val)

    return buf.getvalue()



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
import typing
import uuid

from .transport_device import TransportDevice

@dataclass(frozen=True)
class DeviceAddress:
    """The minimal set of information necessary to communicate with a
    device in a system.  It may be just a CAN ID if that is unique, or
    it may be a UUID prefix.  It may also include a transport device,
    although that is not required."""
    can_id: typing.Optional[int] = None
    uuid: typing.Optional[bytes] = None
    transport_device: typing.Optional[TransportDevice] = None

    def __repr__(self):
        if self.can_id:
            return f'DeviceAddress(can_id={self.can_id}, td={self.transport_device})'
        uuid_bytes = self.uuid.hex() if self.uuid else None
        return f'DeviceAddress(uuid={uuid_bytes}, td={self.transport_device})'


@dataclass
class DeviceInfo:
    """This describes a device that was discovered on the CAN bus.  It
    includes the full available addressing information, as well as the
    minimal DeviceAddress structure necessary to address it in the
    current system."""

    can_id: int = 1
    uuid: typing.Optional[bytes] = None
    transport_device: typing.Optional[TransportDevice] = None
    address: typing.Optional[DeviceAddress] = None

    def __repr__(self):
        uuid_bytes = uuid.UUID(bytes=self.uuid) if self.uuid else None
        return f'DeviceInfo(can_id={self.can_id}, uuid={uuid_bytes}, td={self.transport_device})'

    def _cmp_key(self):
        return (self.can_id, self.uuid or b'')

    def __lt__(self, other):
        if not isinstance(other, DeviceInfo):
            return NotImplemented

        return self._cmp_key() < other._cmp_key()



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import importlib_metadata
import sys
import warnings

from . import fdcanusb_device
from . import pythoncan_device
from . import transport

class FdcanusbFactory:
    PRIORITY = 10

    name = 'fdcanusb'

    def add_args(self, parser):
        try:
            parser.add_argument('--can-disable-brs', action='store_true',
                                help='do not set BRS')
        except argparse.ArgumentError:
            # It must already be set.
            pass

        parser.add_argument('--fdcanusb', type=str, action='append',
                            metavar='FILE',
                            help='path to fdcanusb device')

    def is_args_set(self, args):
        return args and args.fdcanusb

    def __call__(self, args):
        kwargs = {}
        if args and args.can_disable_brs:
            kwargs['disable_brs'] = True

        if args and args.can_debug:
            kwargs['debug_log'] = args.can_debug

        if args and args.fdcanusb:
            return [fdcanusb_device.FdcanusbDevice(path, **kwargs)
                    for path in args.fdcanusb]

        results = [fdcanusb_device.FdcanusbDevice(x, **kwargs) for x in
                   fdcanusb_device.FdcanusbDevice.detect_fdcanusbs()]
        if not results:
            raise RuntimeError('No fdcanusb detected')

        return results


class PythonCanFactory:
    PRIORITY = 11

    name = 'pythoncan'

    def add_args(self, parser):
        try:
            parser.add_argument('--can-disable-brs', action='store_true',
                                help='do not set BRS')
        except argparse.ArgumentError:
            # It must already be set.
            pass
        parser.add_argument('--can-iface', type=str, metavar='IFACE',
                            help='pythoncan "interface" (default: socketcan)')
        parser.add_argument('--can-chan', type=str, action='append',
                            metavar='CHAN',
                            help='pythoncan "channel" (default: can0)')

    def is_args_set(self, args):
        return args and (args.can_iface or args.can_chan)

    def __call__(self, args):
        kwargs = {}
        if args:
            if args.can_iface:
                kwargs['interface'] = args.can_iface
            if args.can_disable_brs:
                kwargs['disable_brs'] = True
            if args.can_debug:
                kwargs['debug_log'] = args.can_debug

        if args and args.can_chan:
            return [
                pythoncan_device.PythonCanDevice(
                    channel=channel, **kwargs)
                for channel in args.can_chan
            ]

        return pythoncan_device.PythonCanDevice.enumerate_devices(**kwargs)


TRANSPORT_FACTORIES = []

_transports_initialized = False

def get_transport_factories():
    global _transports_initialized

    if not _transports_initialized:
        # We initialize these in a deferred manner so that transports
        # are able to import things from moteus as necessary.
        _transports_initialized = True
        TRANSPORT_FACTORIES.extend([
            FdcanusbFactory(),
            PythonCanFactory(),
        ] + [ep.load()() for ep in
             importlib_metadata.entry_points().select(
                 group='moteus.transports2')
             ])

    return TRANSPORT_FACTORIES

GLOBAL_TRANSPORT = None


def make_transport_args(parser):
    for factory in get_transport_factories():
        if hasattr(factory, 'add_args'):
            factory.add_args(parser)

    parser.add_argument(
        '--can-debug', type=str,
        metavar='FILE',
        help='write raw CAN log')
    parser.add_argument(
        '--force-transport', type=str,
        choices=[x.name for x in get_transport_factories()],
        help='Force the given transport type to be used exclusively')


def check_gui_compatibility():
    try:
        import moteus_gui
        from importlib.metadata import version, PackageNotFoundError
        from packaging.version import parse as parse_version

        try:
            gui_version = version('moteus-gui')
            if parse_version(gui_version) < parse_version('0.3.93'):
                warnings.warn(
                    "moteus-gui is outdated.  Please upgrade: python -m pip install -U moteus-gui",
                    UserWarning
                )
        except PackageNotFoundError:
            pass
    except ImportError:
        pass


def get_singleton_transport(args=None):
    global GLOBAL_TRANSPORT

    if GLOBAL_TRANSPORT:
        return GLOBAL_TRANSPORT

    # We check this here because it will likely only be called once
    # per application instance, and will nearly always be called by
    # the GUI.
    check_gui_compatibility()

    if args and args.can_debug:
        args.can_debug = open(args.can_debug, 'wb')

    maybe_result = None
    to_try = sorted(get_transport_factories(), key=lambda x: x.PRIORITY)
    if args and args.force_transport:
        to_try = [x for x in to_try if x.name == args.force_transport]
    elif args:
        # See if any transports have options set.  If so, then limit
        # to just those that do.
        if any([x.is_args_set(args) for x in get_transport_factories()]):
            to_try = [x for x in to_try if x.is_args_set(args)]

    devices = []

    errors = []
    for factory in to_try:
        maybe_results = []
        try:
            maybe_results = factory(args)
        except Exception as e:
            # This can be useful when implementing factories, as the
            # str(e) representation can be a bit abbreviated to find
            # an actual problem in code.
            if False:
                import traceback
                traceback.print_exc()

            errors.append((factory, str(e)))
            pass

        devices.extend(maybe_results)

    if not devices:
        raise RuntimeError("Unable to find a default transport, tried: {}".format(
            ','.join([str(x) for x in errors])))

    GLOBAL_TRANSPORT = transport.Transport(devices)
    return GLOBAL_TRANSPORT



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from .transport_device import Frame, FrameFilter, Subscription
from .transport import Transport


class TransportWrapper:
    """Compatibility base class to implement old-style transports with
    a new TransportDevice."""

    def __init__(self, devices, routing_table={}):
        self._transport = Transport(devices, routing_table=routing_table)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        self._transport.close()

    async def cycle(self, *args, **kwargs):
        return await self._transport.cycle(*args, **kwargs)

    async def write(self, *args, **kwargs):
        return await self._transport.write(*args, **kwargs)

    async def read(self, *args, **kwargs):
        return await self._transport.read(*args, **kwargs)

    async def flush_read_queue(self, *args, **kwargs):
        return await self._transport.flush_read_queue(*args, **kwargs)

    async def discover(self, *args, **kwargs):
        return await self._transport.discover(*args, **kwargs)



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from dataclasses import dataclass, field
import typing


@dataclass
class Frame:
    arbitration_id: int = 0
    data: bytes = b''
    dlc: int = 0
    is_extended_id: bool = False
    is_fd: bool = False
    bitrate_switch: bool = False

    '''Designates which device and sub-channel the frame was received
    on, or should be transmitted to.'''
    channel: typing.Any = None

    # This is a compatibility alias for channel that will be removed
    # in a future version.  It will only be set from calls to
    # 'Transport.cycle' or 'Transport.read'
    bus: typing.Any = None


FrameFilter = typing.Callable[[Frame], bool]


class Subscription:
    '''ABC for managing device subscriptions.'''
    def cancel(self):
        raise NotImplementedError()


class TransportDevice:
    '''Base for hardware-specific CAN devices'''
    def __init__(self, max_buffer_size=50, padding_hex='50'):
        self._max_buffer_size = max_buffer_size
        self._padding_hex = padding_hex

        self._receive_queue = []
        self._receive_waiters = []

        self._subscriptions = {}
        self._next_subscription_id = 0

    def parent(self):
        return None

    def close(self):
        pass

    def empty_bus_tx_safe(self):
        """Return True if it is safe to send frames to this
        TransportDevice if no physical hardware is connected.

        It is implied that if False, then attempts should not be made
        to discover devices attached to it, and broadcast frames
        should not be sent to it.
        """
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    async def send_frame(self, frame: Frame):
        raise NotImplementedError()

    async def receive_frame(self) -> Frame:
        while True:
            if self._receive_queue:
                return self._receive_queue.pop(0)

            try:
                waiter = asyncio.Future()
                self._receive_waiters.append(waiter)

                await waiter
            finally:
                self._receive_waiters = [w for w in self._receive_waiters
                                         if w != waiter]

    @dataclass
    class Request:
        # The frame to send.
        frame: Frame

        # If this "TransportDevice" is a "parent" device with
        # sub-devices, which sub-device to use.
        child_device: typing.Optional[typing.Any] = None

        # Frames that match this filter will be returned.
        frame_filter: typing.Optional[FrameFilter] = None

        expected_reply_size: int = 0

        # Matching frames are returned in this list.
        responses: typing.List[Frame] = field(default_factory=list)

    async def transaction(
            self,
            requests: typing.List[Request]):
        """Send requests and wait for matching responses.

        Results are stored in place in each Request structure.

        If cancelled or interrupted with an asyncio timeout, any
        already received frames will be stored in the Request
        structures.
        """
        raise NotImplementedError()

    def _round_up_dlc(self, size):
        if size <= 8:
            return size
        if size <= 12:
            return 12
        if size <= 16:
            return 16
        if size <= 20:
            return 20
        if size <= 24:
            return 24
        if size <= 32:
            return 32
        if size <= 48:
            return 48
        if size <= 64:
            return 64
        return size

    class _Subscription:
        def __init__(self, parent, subscription_id):
            self.parent = parent
            self.id = subscription_id

        def cancel(self):
            self.parent._subscriptions.pop(self.id, None)

    def _subscribe(self,
                   filter: FrameFilter,
                   handler: typing.Callable[[Frame], typing.Awaitable[None]]) -> Subscription:
        sub_id = self._next_subscription_id
        self._next_subscription_id += 1

        self._subscriptions[sub_id] = (filter, handler)

        return TransportDevice._Subscription(self, sub_id)

    async def _run_handler(self, handler, frame):
        try:
            await handler(frame)
        except Exception as e:
            import traceback, sys
            traceback.print_exc()
            sys.exit(1)

    def _notify_waiters(self, waitlist):
        waiters_to_notify = [w for w in waitlist if not w.done()]
        [w.set_result(None) for w in waiters_to_notify]

    async def _handle_received_frame(self, frame):
        any_subscription = False

        # Then dispatch to subscribers.
        for sub_id, (filter_fn, handler) in self._subscriptions.items():
            if filter_fn(frame):
                any_subscription = True
                asyncio.create_task(self._run_handler(handler, frame))

        if not any_subscription:
            self._receive_queue.append(frame)

            # Limit our maximum queue size.
            while len(self._receive_queue) > self._max_buffer_size:
                self._receive_queue.pop(0)

            self._notify_waiters(self._receive_waiters)



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import argparse
import collections.abc
import copy
import enum
import io
import math
import struct

from . import multiplex as mp
from . import command as cmd
from . import fdcanusb
from . import pythoncan
from .protocol import Register, Mode, Writer, parse_reply, Result, parse_message, make_uuid_prefix
from .transport_factory import TRANSPORT_FACTORIES, get_singleton_transport, make_transport_args

import moteus.reader


def namedtuple_to_dict(obj):
    '''Convert a namedtuple recursively into a nested dictionary.

    This function handles namedtuples, dictionaries, and sequences,
    converting them to nested dictionaries and lists suitable for
    JSON serialization.

    Args:
        obj: The object to convert (typically a namedtuple)

    Returns:
        dict, list, or primitive type representation of the input
    '''

    if isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return {field: namedtuple_to_dict(getattr(obj, field)) for field in obj._fields}

    if isinstance(obj, collections.abc.Mapping):
        return {k: namedtuple_to_dict(v) for k, v in obj.items()}

    if isinstance(obj, collections.abc.Sequence) and not isinstance(obj, (str, bytes, bytearray)):
        return [namedtuple_to_dict(x) for x in obj]

    return obj


def _merge_resolutions(a, b):
    if a == mp.IGNORE:
        return b
    if b == mp.IGNORE:
        return a
    return max(a, b)


class QueryResolution:
    mode = mp.INT8
    position = mp.F32
    velocity = mp.F32
    torque = mp.F32
    q_current = mp.IGNORE
    d_current = mp.IGNORE
    abs_position = mp.IGNORE
    power = mp.IGNORE
    motor_temperature = mp.IGNORE
    trajectory_complete = mp.IGNORE
    rezero_state = mp.IGNORE
    home_state = mp.IGNORE
    voltage = mp.INT8
    temperature = mp.INT8
    fault = mp.INT8

    aux1_gpio = mp.IGNORE
    aux2_gpio = mp.IGNORE

    # Additional registers can be queried by enumerating them as keys
    # in this dictionary, with the resolution as the matching value.
    _extra = {
        # 0x020 : mp.F32, ...
    }


class PositionResolution:
    position = mp.F32
    velocity = mp.F32
    feedforward_torque = mp.F32
    kp_scale = mp.F32
    kd_scale = mp.F32
    maximum_torque = mp.F32
    stop_position = mp.F32
    watchdog_timeout = mp.F32
    velocity_limit = mp.F32
    accel_limit = mp.F32
    fixed_voltage_override = mp.F32
    ilimit_scale = mp.F32
    fixed_current_override = mp.F32
    ignore_position_bounds = mp.F32


class VFOCResolution:
    theta = mp.F32
    voltage = mp.F32
    theta_rate = mp.F32


class CurrentResolution:
    d_A = mp.F32
    q_A = mp.F32


class PwmResolution:
    aux1_pwm1 = mp.INT16
    aux1_pwm2 = mp.INT16
    aux1_pwm3 = mp.INT16
    aux1_pwm4 = mp.INT16
    aux1_pwm5 = mp.INT16
    aux2_pwm1 = mp.INT16
    aux2_pwm2 = mp.INT16
    aux2_pwm3 = mp.INT16
    aux2_pwm4 = mp.INT16
    aux2_pwm5 = mp.INT16


class ZeroVelocityResolution:
    kd_scale = mp.F32


def make_parser(id):
    return parse_message


def parse_diagnostic_message(message, channel):
    data = message.data

    if len(data) < 3:
        return None

    if data[0] != mp.STREAM_SERVER_DATA:
        return None
    if data[1] != channel:
        return None
    datalen, nextoff = mp.read_varuint(2, data)
    if datalen is None:
        return None

    if datalen > (len(data) - nextoff):
        return None
    return data[nextoff:nextoff+datalen]


class DiagnosticResult:
    id = None
    data = b''

    def __repr__(self):
        return f'{self.id}/{self.data}'


def make_diagnostic_parser(channel):
    def parse(message):
        result = DiagnosticResult()
        result.id = (message.arbitration_id >> 8) & 0x7f
        result.data = parse_diagnostic_message(message, channel)
        return result
    return parse


class FaultError(RuntimeError):
    def __init__(self, mode, code):
        super(FaultError, self).__init__(f"Fault mode={mode} code={code}")


class Controller:
    """Operates a single moteus controller across some communication
    medium.

    Attributes:
      id: bus ID of the controller or DeviceAddress structure
      query_resolution: an instance of moteus.QueryResolution
      position_resolution: an instance of moteus.PositionResolution
      transport: something modeling moteus.Transport to send commands through
    """

    def __init__(self, id=1,
                 query_resolution=QueryResolution(),
                 position_resolution=PositionResolution(),
                 vfoc_resolution=VFOCResolution(),
                 current_resolution=CurrentResolution(),
                 pwm_resolution=PwmResolution(),
                 zero_velocity_resolution=ZeroVelocityResolution(),
                 transport=None,
                 source_can_id=0,
                 can_prefix=0x0000):
        self.id = id
        self.source_can_id = source_can_id
        self.query_resolution = query_resolution
        self.position_resolution = position_resolution
        self.vfoc_resolution = vfoc_resolution
        self.current_resolution = current_resolution
        self.pwm_resolution = pwm_resolution
        self.zero_velocity_resolution = zero_velocity_resolution
        self.transport = transport
        self._parser = make_parser(id)
        self._can_prefix = can_prefix

        # Pre-compute our query string.
        self._query_data, self._default_query_reply_size = self._make_query_data()
        self._make_uuid_prefix_data()
        self.max_diagnostic_write = 64 - len(self._uuid_prefix_data) - 3

    def _get_transport(self):
        if self.transport:
            return self.transport

        # Try to construct a global singleton using some agreed upon
        # method that is hookable.
        self.transport = get_singleton_transport()
        return self.transport

    async def flush_transport(self):
        try:
            await asyncio.wait_for(self.transport.read(), 0.02)
        except asyncio.TimeoutError:
            pass

    def _make_uuid_prefix_data(self):
        self._uuid_prefix_data = make_uuid_prefix(self.id)

    def _make_query_data(self, query_resolution=None):
        if query_resolution is None:
            query_resolution = self.query_resolution

        expected_reply_size = 0

        buf = io.BytesIO()
        writer = Writer(buf)
        qr = query_resolution
        c1 = mp.WriteCombiner(writer, 0x10, int(Register.MODE), [
            qr.mode,
            qr.position,
            qr.velocity,
            qr.torque,
            qr.q_current,
            qr.d_current,
            qr.abs_position,
            qr.power,
            ])
        for i in range(c1.size()):
            c1.maybe_write()

        expected_reply_size += c1.reply_size

        c2 = mp.WriteCombiner(writer, 0x10, int(Register.MOTOR_TEMPERATURE), [
            qr.motor_temperature,
            qr.trajectory_complete,
            _merge_resolutions(qr.rezero_state, qr.home_state),
            qr.voltage,
            qr.temperature,
            qr.fault,
        ])
        for i in range(c2.size()):
            c2.maybe_write()

        expected_reply_size += c2.reply_size

        c3 = mp.WriteCombiner(writer, 0x10, int(Register.AUX1_GPIO_STATUS), [
            qr.aux1_gpio,
            qr.aux2_gpio,
        ])
        for i in range(c3.size()):
            c3.maybe_write()

        expected_reply_size += c3.reply_size

        if len(qr._extra):
            min_val = int(min(qr._extra.keys()))
            max_val = int(max(qr._extra.keys()))
            c4 = mp.WriteCombiner(
                writer, 0x10, min_val,
                [qr._extra.get(i, mp.IGNORE)
                 for i in range(min_val, max_val +1)])
            for _ in range(c4.size()):
                c4.maybe_write()
            expected_reply_size += c4.reply_size

        return buf.getvalue(), expected_reply_size

    def _format_query(self, query, query_override, data_buf, result):
        def expect_reply(frame):
            # For a reply to these requests, the first byte should be
            # one of the reply or error frames.
            if len(frame.data) < 1:
                return False

            if frame.data[0] & 0xf0 == 0x20 or frame.data[0] == 0x31:
                return True

            return False

        if query_override is not None:
            query_data, expected_reply_size = \
                self._make_query_data(query_override)
            data_buf.write(query_data)
            result.expected_reply_size = expected_reply_size
            result.reply_filter = expect_reply
        elif query:
            data_buf.write(self._query_data)
            result.expected_reply_size = self._default_query_reply_size
            result.reply_filter = expect_reply

    def _make_command(self, *, query, query_override=None):
        result = cmd.Command()

        result.destination = self.id

        if not isinstance(self.id, int):
            result.channel = self.id.transport_device
            if self.id.can_id is None:
                result.data = self._uuid_prefix_data

        result.source = self.source_can_id
        result.reply_required = query or (query_override is not None)
        result.parse = self._parser
        result.can_prefix = self._can_prefix
        result.expected_reply_size = self._default_query_reply_size if query else 0

        # Create and properly position BytesIO for data construction
        data_buf = io.BytesIO(result.data if result.data else b'')
        data_buf.seek(0, io.SEEK_END)

        return result, data_buf

    def make_query(self, query_override=None):
        result, _ = self._make_command(
            query=True, query_override=query_override)
        if query_override:
            result.data, result.expected_reply_size = \
                self._make_query_data(query_override)
        else:
            result.data = self._query_data
            result.expected_reply_size = self._default_query_reply_size

        def expect_reply(frame):
            if len(frame.data) < 1:
                return False

            if frame.data[0] & 0xf0 == 0x20 or frame.data[0] == 0x31:
                return True

            return False

        result.reply_filter = expect_reply
        return result

    async def query(self, **kwargs):
        return await self.execute(self.make_query(**kwargs))

    def make_custom_query(self, to_query_fields):
        """Return a moteus.Command structure with data required to query the
        registers given by the 'to_query_fields' dictionary of
        registers to resolutions.
        """

        result, data_buf = self._make_command(query=True)

        writer = Writer(data_buf)

        min_val = int(min(to_query_fields.keys()))
        max_val = int(max(to_query_fields.keys()))
        c = mp.WriteCombiner(writer, 0x10, min_val,
                             [to_query_fields.get(i, mp.IGNORE)
                              for i in range(min_val, max_val + 1)])
        for _ in range(min_val, max_val + 1):
            c.maybe_write()

        result.data = data_buf.getvalue()
        result.expected_reply_size = c.reply_size

        def expect_reply(frame):
            if len(frame.data) < 1:
                return False

            if frame.data[0] & 0xf0 == 0x20 or frame.data[0] == 0x31:
                return True

            return False

        result.reply_filter = expect_reply
        return result

    async def custom_query(self, *args, **kwargs):
        return await self.execute(self.make_custom_query(*args, **kwargs))

    def make_stop(self, *, query=False, query_override=None):
        """Return a moteus.Command structure with data necessary to send a
        stop mode command."""

        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.STOPPED))

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_stop(self, *args, **kwargs):
        return await self.execute(self.make_stop(**kwargs))

    def make_zero_velocity(self,
                           *,
                           kd_scale=None,
                           query=False,
                           query_override=None):
        """Return a moteus.Command structure with data necessary to send a
        zero velocity mode command."""

        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        zr = self.zero_velocity_resolution
        resolutions = [
            zr.kd_scale if kd_scale is not None else mp.IGNORE,
        ]

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.ZERO_VELOCITY))

        # Only write kd_scale if it's not None
        if kd_scale is not None:
            combiner = mp.WriteCombiner(
                writer, 0x00, int(Register.COMMAND_KD_SCALE), resolutions)

            if combiner.maybe_write():
                writer.write_pwm(kd_scale, zr.kd_scale)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_zero_velocity(self, *args, **kwargs):
        return await self.execute(self.make_zero_velocity(**kwargs))

    def make_set_output(self, *args,
                        position=0.0,
                        query=False,
                        query_override=None,
                        cmd=None
    ):
        """Return a moteus.Command structure with data necessary to send a
        set output nearest command."""

        if len(args):
            raise ValueError(f'unexpected positional arguments: {args}')

        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_F32 | 0x01)
        writer.write_varuint(cmd)
        writer.write_f32(position)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()
        return result

    def make_set_output_nearest(self, *args,
                                position=0.0,
                                query=False,
                                query_override=None):
        return self.make_set_output(
            *args,
            position=position, query=query, query_override=query_override,
            cmd=Register.SET_OUTPUT_NEAREST)

    def make_set_output_exact(self, *args,
                              position=0.0,
                              query=False,
                              query_override=None):
        return self.make_set_output(
            *args,
            position=position, query=query, query_override=query_override,
            cmd=Register.SET_OUTPUT_EXACT)

    async def set_output(self, *args, cmd=None, **kwargs):
        return await self.execute(self.make_set_output(*args, **kwargs, cmd=cmd))

    async def set_output_nearest(self, *args, **kwargs):
        return await self.set_output(*args, cmd=Register.SET_OUTPUT_NEAREST, **kwargs)

    async def set_output_exact(self, *args, **kwargs):
        return await self.set_output(*args, cmd=Register.SET_OUTPUT_EXACT, **kwargs)


    # For backwards compatibility, "*_output_nearest" used to be named
    # "make/set_rezero".
    def make_rezero(self, *args,
                    rezero=0.0,
                    query=False,
                    query_override=None):
        return self.make_set_output(
            *args,
            position=rezero, query=query, query_override=query_override,
            cmd=Register.SET_OUTPUT_NEAREST)

    async def set_rezero(self, *args, **kwargs):
        return await self.execute(self.make_rezero(**kwargs))

    def make_require_reindex(self,
                             query=False,
                             query_override=None):
        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_varuint(Register.REQUIRE_REINDEX)
        writer.write_int8(1)

        result.data = data_buf.getvalue()
        return result

    async def set_require_reindex(self, query=False, query_override=None):
        return await self.execute(self.make_require_reindex(
            query=query, query_override=query_override))

    def make_recapture_position_velocity(self,
                                         query=False,
                                         query_override=None):
        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_varuint(Register.RECAPTURE_POSITION_VELOCITY)
        writer.write_int8(1)

        result.data = data_buf.getvalue()
        return result

    async def set_recapture_position_velocity(self,
                                              query=False,
                                              query_override=None):
        return await self.execute(self.make_recapture_position_velocity(
            query=query, query_override=query_override))

    def make_position(self,
                      *,
                      position=None,
                      velocity=None,
                      feedforward_torque=None,
                      kp_scale=None,
                      kd_scale=None,
                      maximum_torque=None,
                      stop_position=None,
                      watchdog_timeout=None,
                      velocity_limit=None,
                      accel_limit=None,
                      fixed_voltage_override=None,
                      ilimit_scale=None,
                      fixed_current_override=None,
                      ignore_position_bounds=None,
                      query=False,
                      query_override=None):
        """Return a moteus.Command structure with data necessary to send a
        position mode command with the given values."""

        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        pr = self.position_resolution
        resolutions = [
            pr.position if position is not None else mp.IGNORE,
            pr.velocity if velocity is not None else mp.IGNORE,
            pr.feedforward_torque if feedforward_torque is not None else mp.IGNORE,
            pr.kp_scale if kp_scale is not None else mp.IGNORE,
            pr.kd_scale if kd_scale is not None else mp.IGNORE,
            pr.maximum_torque if maximum_torque is not None else mp.IGNORE,
            pr.stop_position if stop_position is not None else mp.IGNORE,
            pr.watchdog_timeout if watchdog_timeout is not None else mp.IGNORE,
            pr.velocity_limit if velocity_limit is not None else mp.IGNORE,
            pr.accel_limit if accel_limit is not None else mp.IGNORE,
            pr.fixed_voltage_override if fixed_voltage_override is not None else mp.IGNORE,
            pr.ilimit_scale if ilimit_scale is not None else mp.IGNORE,
            pr.fixed_current_override if fixed_current_override is not None else mp.IGNORE,
            pr.ignore_position_bounds if ignore_position_bounds is not None else mp.IGNORE,
        ]

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.POSITION))

        combiner = mp.WriteCombiner(
            writer, 0x00, int(Register.COMMAND_POSITION), resolutions)

        if combiner.maybe_write():
            writer.write_position(position, pr.position)
        if combiner.maybe_write():
            writer.write_velocity(velocity, pr.velocity)
        if combiner.maybe_write():
            writer.write_torque(feedforward_torque, pr.feedforward_torque)
        if combiner.maybe_write():
            writer.write_pwm(kp_scale, pr.kp_scale)
        if combiner.maybe_write():
            writer.write_pwm(kd_scale, pr.kd_scale)
        if combiner.maybe_write():
            writer.write_torque(maximum_torque, pr.maximum_torque)
        if combiner.maybe_write():
            writer.write_position(stop_position, pr.stop_position)
        if combiner.maybe_write():
            writer.write_time(watchdog_timeout, pr.watchdog_timeout)
        if combiner.maybe_write():
            writer.write_velocity(velocity_limit, pr.velocity_limit)
        if combiner.maybe_write():
            writer.write_accel(accel_limit, pr.accel_limit)
        if combiner.maybe_write():
            writer.write_voltage(fixed_voltage_override, pr.fixed_voltage_override)
        if combiner.maybe_write():
            writer.write_pwm(ilimit_scale, pr.ilimit_scale)
        if combiner.maybe_write():
            writer.write_current(fixed_current_override, pr.fixed_current_override)
        if combiner.maybe_write():
            writer.write_int(ignore_position_bounds, pr.ignore_position_bounds)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_position(self, *args, **kwargs):
        return await self.execute(self.make_position(**kwargs))

    async def set_position_wait_complete(
            self,
            period_s=0.025,
            query_override=None,
            *args, **kwargs):
        """Repeatedly send a position mode command to a device until it
        reports that the trajectory has been completed.

        If the controller is unresponsive, this method will never return.

        If the controller reports a fault or position mode timeout, a
        FaultError exception will be raised.
        """

        if query_override is None:
            query_override = copy.deepcopy(self.query_resolution)
        else:
            query_override = copy.deepcopy(query_override)

        if query_override.mode == mp.IGNORE:
            query_override.mode = mp.INT8
        if query_override.fault == mp.IGNORE:
            query_override.fault = mp.INT8
        query_override.trajectory_complete = mp.INT8

        count = 2
        while True:
            result = await self.set_position(
                query_override=query_override, *args, **kwargs)

            if result is not None:
                count = max(count - 1, 0)

            if (count == 0 and
                result is not None and
                result.values[Register.TRAJECTORY_COMPLETE]):
                return result

            current_mode = result.values.get(Register.MODE, Mode.STOPPED)
            fault_code = result.values.get(Register.FAULT, 0)

            if current_mode == Mode.FAULT or current_mode == Mode.TIMEOUT:
                raise FaultError(current_mode, fault_code)

            await asyncio.sleep(period_s)

    def make_vfoc(self,
                  *,
                  theta,
                  voltage,
                  theta_rate=0.0,
                  query=False,
                  query_override=None):
        """Return a moteus.Command structure with data necessary to send a
        voltage mode FOC command."""

        result, data_buf = self._make_command(
            query=query, query_override=query_override)
        cr = self.vfoc_resolution
        resolutions = [
            cr.theta if theta is not None else mp.IGNORE,
            cr.voltage if voltage is not None else mp.IGNORE,
            mp.IGNORE,
            mp.IGNORE,
            mp.IGNORE,
            mp.IGNORE,
            cr.theta_rate if (theta_rate != 0.0 and theta_rate is not None) else mp.IGNORE,
        ]

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.VOLTAGE_FOC))

        combiner = mp.WriteCombiner(
            writer, 0x00, int(Register.VFOC_THETA), resolutions)

        if combiner.maybe_write():
            writer.write_pwm(theta / math.pi, cr.theta)
        if combiner.maybe_write():
            writer.write_voltage(voltage, cr.voltage)
        if combiner.maybe_write():
            assert False
        if combiner.maybe_write():
            assert False
        if combiner.maybe_write():
            assert False
        if combiner.maybe_write():
            assert False
        if combiner.maybe_write():
            writer.write_velocity(theta_rate / math.pi, cr.theta_rate)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_vfoc(self, *args, **kwargs):
        return await self.execute(self.make_vfoc(**kwargs))

    def make_current(self,
                     *,
                     d_A,
                     q_A,
                     query=False,
                     query_override=None):
        """Return a moteus.Command structure with data necessary to send a
        current mode command.
        """

        result, data_buf = self._make_command(
            query=query, query_override=query_override)
        cr = self.current_resolution
        resolutions = [
            cr.d_A if d_A is not None else mp.IGNORE,
            cr.q_A if q_A is not None else mp.IGNORE,
        ]

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.CURRENT))

        # Yes, annoyingly the register mapping as of version 4 still
        # has the Q current first in this grouping, unlike everywhere
        # else where D current is first.
        combiner = mp.WriteCombiner(
            writer, 0x00, int(Register.COMMAND_Q_CURRENT), resolutions)

        if combiner.maybe_write():
            writer.write_current(q_A, cr.q_A)
        if combiner.maybe_write():
            writer.write_current(d_A, cr.d_A)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_current(self, *args, **kwargs):
        return await self.execute(self.make_current(**kwargs))

    def make_stay_within(
            self,
            *,
            lower_bound=None,
            upper_bound=None,
            feedforward_torque=None,
            kp_scale=None,
            kd_scale=None,
            maximum_torque=None,
            stop_position=None,
            watchdog_timeout=None,
            ilimit_scale=None,
            ignore_position_bounds=None,
            query=False,
            query_override=None):
        """Return a moteus.Command structure with data necessary to send a
        within mode command with the given values."""

        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        pr = self.position_resolution
        resolutions = [
            pr.position if lower_bound is not None else mp.IGNORE,
            pr.position if upper_bound is not None else mp.IGNORE,
            pr.feedforward_torque if feedforward_torque is not None else mp.IGNORE,
            pr.kp_scale if kp_scale is not None else mp.IGNORE,
            pr.kd_scale if kd_scale is not None else mp.IGNORE,
            pr.maximum_torque if maximum_torque is not None else mp.IGNORE,
            pr.watchdog_timeout if watchdog_timeout is not None else mp.IGNORE,
            pr.ilimit_scale if ilimit_scale is not None else mp.IGNORE,
            pr.ignore_position_bounds if ignore_position_bounds is not None else mp.IGNORE,
        ]

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.STAY_WITHIN))

        combiner = mp.WriteCombiner(
            writer, 0x00, int(Register.COMMAND_WITHIN_LOWER_BOUND),
            resolutions)

        if combiner.maybe_write():
            writer.write_position(lower_bound, pr.position)
        if combiner.maybe_write():
            writer.write_position(upper_bound, pr.position)
        if combiner.maybe_write():
            writer.write_torque(feedforward_torque, pr.feedforward_torque)
        if combiner.maybe_write():
            writer.write_pwm(kp_scale, pr.kp_scale)
        if combiner.maybe_write():
            writer.write_pwm(kd_scale, pr.kd_scale)
        if combiner.maybe_write():
            writer.write_torque(maximum_torque, pr.maximum_torque)
        if combiner.maybe_write():
            writer.write_time(watchdog_timeout, pr.watchdog_timeout)
        if combiner.maybe_write():
            writer.write_pwm(ilimit_scale, pr.ilimit_scale)
        if combiner.maybe_write():
            writer.write_int(ignore_position_bounds, pr.ignore_position_bounds)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_stay_within(self, *args, **kwargs):
        return await self.execute(self.make_stay_within(**kwargs))

    def make_brake(self, *, query=False, query_override=None):
        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT8 | 0x01)
        writer.write_int8(int(Register.MODE))
        writer.write_int8(int(Mode.BRAKE))

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_brake(self, *args, **kwargs):
        return await self.execute(self.make_brake(**kwargs))

    def make_write_gpio(self, aux1=None, aux2=None,
                        query=False, query_override=None):
        """Return a moteus.Command structure with data necessary to set one or
        more GPIO registers.

        aux1/aux2 are an optional integer bitfield, where the least
        significant bit is pin 0 on the respective port.
        """

        result, data_buf = self._make_command(
            query=query, query_override=query_override)

        writer = Writer(data_buf)

        combiner = mp.WriteCombiner(
            writer, 0x00, int(Register.AUX1_GPIO_COMMAND), [
                mp.INT8 if aux1 is not None else mp.IGNORE,
                mp.INT8 if aux2 is not None else mp.IGNORE,
        ])

        if combiner.maybe_write():
            writer.write_int8(aux1)
        if combiner.maybe_write():
            writer.write_int8(aux2)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()
        return result

    async def set_write_gpio(self, *args, **kwargs):
        return await self.execute(self.make_write_gpio(**kwargs))

    def make_read_gpio(self):
        """Return a moteus.Command structure with data necessary to read all
        GPIO digital inputs."""

        result, data_buf = self._make_command(query=True)
        writer = Writer(data_buf)

        combiner = mp.WriteCombiner(
            writer, 0x10, int(Register.AUX1_GPIO_STATUS), [
                mp.INT8,
                mp.INT8,
        ])

        for i in range(2):
            combiner.maybe_write()

        result.data = data_buf.getvalue()
        return result

    async def read_gpio(self):
        """Return a bytes() object with an int8 for each auxiliary port.  The
        pins for each port are represented as bits, with the least significant
        bit being pin 0.

        None can be returned if no response is received.
        """

        results = await self._get_transport().cycle([self.make_read_gpio()])
        if len(results) == 0:
            return None
        result = results[0]
        return bytes([result.values[Register.AUX1_GPIO_STATUS],
                      result.values[Register.AUX2_GPIO_STATUS]])

    def make_diagnostic_write(self, data, channel=1):
        result, data_buf = self._make_command(query=False)

        # CAN-FD frames can be at most 64 bytes long
        assert len(data) <= 61

        writer = Writer(data_buf)
        writer.write_int8(mp.STREAM_CLIENT_DATA)
        writer.write_int8(channel)  # channel
        writer.write_int8(len(data))
        data_buf.write(data)

        result.data = data_buf.getvalue()

        return result

    async def send_diagnostic_write(self, *args, **kwargs):
        await self._get_transport().cycle([self.make_diagnostic_write(**kwargs)])

    def make_diagnostic_read(self, max_length=48, channel=1):
        result, data_buf = self._make_command(query=True)

        writer = Writer(data_buf)
        writer.write_int8(mp.STREAM_CLIENT_POLL)
        writer.write_int8(channel)
        writer.write_int8(max_length)

        result.parse = make_diagnostic_parser(channel)

        result.data = data_buf.getvalue()
        result.expected_reply_size = 3 + max_length

        def expect_diagnostic_response(frame):
            if len(frame.data) < 3:
                return False

            return frame.data[0] == 0x41

        result.reply_filter = expect_diagnostic_response

        return result

    async def diagnostic_read(self, *args, **kwargs):
        return await self._get_transport().cycle(
            [self.make_diagnostic_read(**kwargs)])

    def make_set_trim(self, *, trim=0):
        result, data_buf = self._make_command(query=False)

        writer = Writer(data_buf)
        writer.write_int8(mp.WRITE_INT32 | 0x01)
        writer.write_varuint(Register.CLOCK_TRIM)
        writer.write_int32(trim)

        result.data = data_buf.getvalue()
        return result

    async def set_trim(self, *args, **kwargs):
        return await self.execute(self.make_set_trim(*args, **kwargs))

    def make_aux_pwm(self, *,
                     aux1_pwm1=None,
                     aux1_pwm2=None,
                     aux1_pwm3=None,
                     aux1_pwm4=None,
                     aux1_pwm5=None,
                     aux2_pwm1=None,
                     aux2_pwm2=None,
                     aux2_pwm3=None,
                     aux2_pwm4=None,
                     aux2_pwm5=None,
                     query=False,
                     query_override=None):
        result, data_buf = self._make_command(query=query, query_override=query_override)

        pr = self.pwm_resolution
        resolutions = [
            pr.aux1_pwm1 if aux1_pwm1 is not None else mp.IGNORE,
            pr.aux1_pwm2 if aux1_pwm2 is not None else mp.IGNORE,
            pr.aux1_pwm3 if aux1_pwm3 is not None else mp.IGNORE,
            pr.aux1_pwm4 if aux1_pwm4 is not None else mp.IGNORE,
            pr.aux1_pwm5 if aux1_pwm5 is not None else mp.IGNORE,
            pr.aux2_pwm1 if aux2_pwm1 is not None else mp.IGNORE,
            pr.aux2_pwm2 if aux2_pwm2 is not None else mp.IGNORE,
            pr.aux2_pwm3 if aux2_pwm3 is not None else mp.IGNORE,
            pr.aux2_pwm4 if aux2_pwm4 is not None else mp.IGNORE,
            pr.aux2_pwm5 if aux2_pwm5 is not None else mp.IGNORE,
        ]

        writer = Writer(data_buf)
        combiner = mp.WriteCombiner(
            writer, 0x00, int(Register.AUX1_PWM1), resolutions)

        if combiner.maybe_write():
            writer.write_pwm(aux1_pwm1, pr.aux1_pwm1)
        if combiner.maybe_write():
            writer.write_pwm(aux1_pwm2, pr.aux1_pwm2)
        if combiner.maybe_write():
            writer.write_pwm(aux1_pwm3, pr.aux1_pwm3)
        if combiner.maybe_write():
            writer.write_pwm(aux1_pwm4, pr.aux1_pwm4)
        if combiner.maybe_write():
            writer.write_pwm(aux1_pwm5, pr.aux1_pwm5)
        if combiner.maybe_write():
            writer.write_pwm(aux2_pwm1, pr.aux2_pwm1)
        if combiner.maybe_write():
            writer.write_pwm(aux2_pwm2, pr.aux2_pwm2)
        if combiner.maybe_write():
            writer.write_pwm(aux2_pwm3, pr.aux2_pwm3)
        if combiner.maybe_write():
            writer.write_pwm(aux2_pwm4, pr.aux2_pwm4)
        if combiner.maybe_write():
            writer.write_pwm(aux2_pwm5, pr.aux2_pwm5)

        self._format_query(query, query_override, data_buf, result)

        result.data = data_buf.getvalue()

        return result

    async def set_aux_pwm(self, *args, **kwargs):
        return await self.execute(self.make_aux_pwm(*args, **kwargs))

    def _extract(self, value):
        if len(value):
            return value[0]
        return None

    async def execute(self, command):
        return self._extract(await self._get_transport().cycle([command]))


class CommandError(RuntimeError):
    def __init__(self, message):
        super(CommandError, self).__init__("Error response:" + message)
        self.message = message


class Stream:
    """Presents a python file-like interface to the diagnostic stream of a
    moteus controller."""

    def __init__(self, controller, verbose=False, channel=1):
        self.controller = controller
        self.verbose = verbose
        self.channel = channel

        self.lock = asyncio.Lock()
        self._read_data = b''
        self._write_data = b''

        self._readers = {}
        self._maxlen = controller.max_diagnostic_write

    def write(self, data):
        self._write_data += data

    async def drain(self):
        while len(self._write_data):
            to_write, self._write_data = self._write_data[0:self._maxlen], self._write_data[self._maxlen:]

            async with self.lock:
                await self.controller.send_diagnostic_write(
                    data=to_write, channel=self.channel)

    async def read(self, size, block=True):
        while ((block == True and len(self._read_data) < size)
               or len(self._read_data) == 0):
            bytes_to_request = min(self._maxlen, size - len(self._read_data))

            async with self.lock:
                these_results = await self.controller.diagnostic_read(
                    bytes_to_request, channel=self.channel)

            this_data = b''.join(x.data for x in these_results if x.data)

            self._read_data += this_data

            if len(this_data) == 0:
                # Wait a bit before asking again.
                await asyncio.sleep(0.01)

        to_return, self._read_data = self._read_data[0:size], self._read_data[size:]
        return to_return

    async def flush_read(self, timeout=0.2):
        self._read_data = b''

        try:
            await asyncio.wait_for(self.read(65536), timeout)
            raise RuntimeError("More data to flush than expected")
        except asyncio.TimeoutError:
            # This is the expected path.
            pass

        self._read_data = b''

        # Now flush anything from the underlying transport if applicable.
        await self.controller.flush_transport()

    async def _read_maybe_empty_line(self):
        while b'\n' not in self._read_data and b'\r' not in self._read_data:
            async with self.lock:
                these_results = await self.controller.diagnostic_read(
                    61, channel=self.channel)

            this_data = b''.join(x.data for x in these_results if x.data)

            self._read_data += this_data

            if len(this_data) == 0:
                await asyncio.sleep(0.01)

        first_newline = min((self._read_data.find(c) for c in b'\r\n'
                             if c in self._read_data), default=None)
        to_return, self._read_data = (
            self._read_data[0:first_newline+1],
            self._read_data[first_newline+1:])
        return to_return

    async def readline(self):
        while True:
            line = (await self._read_maybe_empty_line()).rstrip()
            if len(line) > 0:
                if self.verbose:
                    print(f"< {line}")
                return line

    async def read_until_OK(self):
        result = b''
        while True:
            line = await self.readline()
            if line.startswith(b'OK'):
                return result
            if line.startswith(b'ERR'):
                raise CommandError(line.decode('latin1'))
            result += (line + b'\n')

    async def command(self, data, allow_any_response=False):
        await self.write_message(data)

        if allow_any_response:
            result = await self.readline()
        else:
            result = await self.read_until_OK()
        return result

    async def write_message(self, data):
        if self.verbose:
            print(f"> {data}")

        self.write(data + b'\n')
        await self.drain()

    async def read_binary_blob(self):
        size_bytes = await self.read(5, block=True)
        if size_bytes[0] != 0x0a:
            raise RuntimeError("missing newline before blob")
        size, = struct.unpack('<I', size_bytes[1:])
        return await self.read(size, block=True)

    async def read_data(self, name):
        if name not in self._readers:
            await self.write_message(f"tel schema {name}".encode('latin1'))

            maybe_schema_announce = await self.readline()
            if maybe_schema_announce != f"schema {name}".encode('latin1'):
                raise RuntimeError(
                    f"Unexpected schema announce for '{name}' " +
                    f": '{maybe_schema_announce}'")

            schema = await self.read_binary_blob()
            self._readers[name] = moteus.reader.Type.from_binary(io.BytesIO(schema))

            # Set this to be emitted as binary
            await self.command(f"tel fmt {name} 0".encode('latin1'))

        reader = self._readers[name]
        await self.write_message(f"tel get {name}".encode('latin1'))
        maybe_data_announce = await self.readline()

        if maybe_data_announce != f"emit {name}".encode('latin1'):
            raise RuntimeError(
                f"Invalid data announce for '{name}' : " +
                f"'{maybe_data_announce}'")

        data = await self.read_binary_blob()
        return reader.read(moteus.reader.Stream(io.BytesIO(data)))



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import collections
import io
import itertools
import struct
import typing

from .device_info import DeviceAddress, DeviceInfo
from .protocol import Register, Mode, parse_message, Writer, make_uuid_prefix
from .transport_device import Frame, TransportDevice

from . import command
from . import multiplex as mp


def _make_uuid_query():
    """Create a query command to read the device UUID.

    This creates a query for UUID1-4 registers which contain the full 16-byte
    device UUID (4 bytes per register).

    Returns:
        Tuple of (query_data_bytes, expected_reply_size)
    """
    buf = io.BytesIO()
    writer = Writer(buf)

    # Create WriteCombiner for UUID registers (0x150-0x153)
    # All UUID registers are read as INT32
    uuid_combiner = mp.WriteCombiner(writer, 0x10, int(Register.UUID1), [
        mp.INT32,  # UUID1
        mp.INT32,  # UUID2
        mp.INT32,  # UUID3
        mp.INT32,  # UUID4
        mp.IGNORE, # UUID1_MASK
        mp.IGNORE, # UUID2_MASK
        mp.IGNORE, # UUID3_MASK
        mp.IGNORE, # UUID4_MASK
        mp.INT8,   # UUID_MASK_CAPABLE
    ])

    expected_reply_size = 0
    for i in range(uuid_combiner.size()):
        uuid_combiner.maybe_write()

    expected_reply_size += uuid_combiner.reply_size

    return buf.getvalue(), expected_reply_size


def _extract_uuid_from_result(result):
    """Extract UUID bytes from a query result containing UUID1-4.

    Args:
        result: Query result containing UUID1-4 register values

    Returns:
        16-byte UUID as bytes, or None if UUID registers not present
        or UUID mask not capable

    """
    if not hasattr(result, 'values') or not result.values:
        return None

    values = result.values

    # Check if all UUID registers are present
    uuid_regs = [Register.UUID1, Register.UUID2, Register.UUID3, Register.UUID4, Register.UUID_MASK_CAPABLE]
    if not all(reg in values for reg in uuid_regs):
        return None

    if values[Register.UUID_MASK_CAPABLE] == 0:
        return None

    # Combine the 4 32-bit values into a 16-byte UUID
    # Each register is stored as little-endian 32-bit integer
    uuid_bytes = b''
    for reg in uuid_regs[0:4]:
        # Convert signed int to unsigned int for packing
        val = int(values[reg])
        if val < 0:
            val = val + (1 << 32)  # Convert signed to unsigned 32-bit
        uuid_bytes += struct.pack('<I', val)

    return uuid_bytes


class Transport:
    """This is an object which can dispatch commands to one or more
    controllers across potentially multiple CAN-FD busses.
    """

    def __init__(self, devices, routing_table={}):
        """Initialize the Transport with one or more devices.

        Args:
          devices: Either a single TransportDevice or a list of instances.

          routing_table: A dictionary from DeviceAddress ->
            TransportDevice.  If an addressed DeviceAddress is not
            present, and multiple TransportDevices are used, then
            discovery is performed to find the correct TransportDevice
            upon first use.

        """
        if isinstance(devices, list):
            self._devices = devices
        else:
            self._devices = [devices]

        self._parent_devices = list(set(
            [d.parent() or d for d in self._devices]))

        # This is necessary for some compatibility hacks down below.
        self._pi3hat_device = next(iter(
            [x for x in (self._devices + self._parent_devices)
             if type(x).__name__ == 'Pi3HatDevice']),
            None)

        if (self._pi3hat_device and
            self._pi3hat_device.servo_bus_map and
            not routing_table):
            # Attempt to construct an appropriate routing table.
            children = {
                d.bus() : d
                for d in self._devices
                if type(d).__name__ == 'Pi3HatChildDevice'
            }
            self._routing_table = {}
            for bus, id_list in self._pi3hat_device.servo_bus_map.items():
                for id in id_list:
                    self._routing_table[DeviceAddress(can_id=id)] = children[bus]
        else:
            self._routing_table = routing_table

        # This will hold frames that were received, but we could not
        # yet deliver them to the client because the calls were
        # cancelled.  They will be delivered during calls to .read(),
        # if any are made.
        self._cancelled_queue = []

        # So that the queue does not grow without bound, we limit it
        # to this many frames.
        self._cancel_queue_max_size = 100


    def devices(self):
        return self._devices

    def count(self):
        return len(self._devices)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _add_cancelled_frames(self, to_add):
        self._cancelled_queue.extend(to_add)
        while len(self._cancelled_queue) > self._cancel_queue_max_size:
            self._cancel_queue_max_size.pop(0)

    def close(self):
        for device in self._devices:
            device.close()

    async def _read_frames_until_timeout(self, device, end_time):
        results = []
        while True:
            try:
                remaining = end_time - asyncio.get_event_loop().time()

                if remaining <= 0:
                    break

                response = await asyncio.wait_for(
                    device.receive_frame(),
                    remaining)

                results.append(response)
            except asyncio.TimeoutError:
                break
        return results

    async def _get_devices_for_command(self, cmd, timeout=0.1):
        if len(self._devices) == 1:
            return self._devices

        channel = getattr(cmd, 'channel', None)
        if channel is not None:
            return [channel]

        # For legacy support, we may have a 'bus' attribute on the
        # command.
        if getattr(cmd, 'bus', None) is not None:
            return [d for d in self._devices
                    if hasattr(d, 'bus') and d.bus() == cmd.bus]

        # Make a DeviceAddress version to use everywhere here.
        cmd_destination = (
            cmd.destination
            if not isinstance(cmd.destination, int) else
            DeviceAddress(can_id=cmd.destination))

        # If this is a broadcast request, then we are going to send to
        # everything.
        if (cmd_destination.can_id == 0x7f):
            return [d for d in self._devices
                    if d.empty_bus_tx_safe()]

        # We have multiple devices, and we don't know which one has
        # this destination.  First check the cache.

        maybe_device = self._routing_table.get(cmd_destination, None)
        if maybe_device is not None:
            return [maybe_device]

        # Emit a simple time-limited request on all supported
        # interfaces to see where it responds and save that result.
        discover_command = command.Command()
        discover_command.destination = cmd_destination
        discover_command.source = cmd.source
        discover_command.reply_required = True
        discover_command.can_prefix = cmd.can_prefix
        discover_command.arbitration_id = None

        if (cmd_destination.can_id is None):
            # We need to request only this UUID to respond.
            discover_command.data = make_uuid_prefix(cmd_destination)

        discover_command.data += bytes([0x50])

        # Flush everything on all transports.
        await self.flush_read_queue()

        # Send the frames.
        frame = self._command_to_frame(discover_command)
        for device in self._devices:
            if not device.empty_bus_tx_safe():
                continue
            await device.send_frame(frame)

        # Read everything.  If the system is consistent, we should
        # only get one response across all devices.
        start_time = asyncio.get_event_loop().time()
        end_time = start_time + timeout
        tasks = [
            self._read_frames_until_timeout(device, end_time)
            for device in self._devices
            if device.empty_bus_tx_safe()
        ]
        result_lists = await asyncio.gather(*tasks)
        results = []
        for single_list in result_lists:
            results.extend(single_list)

        # Ideally, we have a single result here.
        if len(results) < 1:
            raise RuntimeError(f'{cmd_destination} not found on any CAN bus')
        if len(results) > 1:
            raise RuntimeError(f'More than one {cmd_destination} found across connected CAN busses {results}')

        self._routing_table[cmd_destination] = results[0].channel
        return [results[0].channel]

    def _make_canid(self, device_address):
        if isinstance(device_address, int):
            return device_address
        if device_address.can_id is not None:
            return device_address.can_id
        return 0x7f

    def _command_to_frame(self, command) -> Frame:
        if getattr(command, 'raw', False):
            return Frame(
                arbitration_id=command.arbitration_id,
                data=command.data,
                is_extended_id=command.arbitration_id > 0x7ff,
                is_fd=True,
                bitrate_switch=True,
                channel=command.channel)

        arbitration_id = (
            self._make_canid(command.destination) |
            (0x8000 if command.reply_required else 0) |
            (command.source << 8) |
            (command.can_prefix << 16))

        return Frame(
            arbitration_id=arbitration_id,
            data=command.data,
            is_extended_id=arbitration_id > 0x7ff,
            is_fd=True,
            bitrate_switch=True)

    def _fix_result_frame(self, frame):
        if getattr(frame, 'channel', None):
            frame.bus = frame.channel

        return frame

    def _make_response_filter(self, command):
        if getattr(command, 'raw', False):
            # For raw commands, accept any response.  In practice,
            # that means that they can't be intermingled with moteus
            # responses.
            return lambda f: True

        dest_id = self._make_canid(command.destination)
        source_id = command.source
        reply_filter = getattr(command, 'reply_filter', None)
        prefix = getattr(command, 'can_prefix', 0)

        def predicate(f,
                      dest_id=dest_id,
                      source_id=source_id,
                      prefix=prefix):
            if ((f.arbitration_id >> 16) & 0x1fff) != prefix:
                return False

            if ((f.arbitration_id >> 8) & 0x7f != dest_id and
                dest_id != 0x7f):
                return False

            if f.arbitration_id & 0x7f != source_id:
                return False

            if reply_filter and not reply_filter(f):
                return False

            return True

        return predicate

    async def _cycle_batch(self, commands, request_attitude,
                           read_unsolicited, force_can_check):
        # Group requests by device.  This is a map from device to:
        #  [TransportDevice.Request, ...]
        device_requests = collections.defaultdict(list)

        for i, command in enumerate(commands):
            frame = self._command_to_frame(command)
            reply_required = command.reply_required
            response_filter = (self._make_response_filter(command)
                               if reply_required else None)

            devices = await self._get_devices_for_command(command)
            for device in devices:
                maybe_parent = device.parent()

                to_send_device = maybe_parent or device

                request = TransportDevice.Request(
                    frame=frame,
                    child_device=device if maybe_parent else None,
                    frame_filter=response_filter,
                    expected_reply_size=command.expected_reply_size)
                request.command = command

                device_requests[to_send_device].append(request)

        if read_unsolicited:
            for device in read_unsolicited:
                maybe_parent = device.parent()
                to_send_device = maybe_parent or device

                device_requests[to_send_device].append(
                    TransportDevice.Request(
                        frame=None,
                        child_device=device if maybe_parent else None,
                        frame_filter=lambda x: True))

        if request_attitude and self._pi3hat_device:
            # This is a compatibility/performance hack that only works
            # with the pi3hat.  Logically we shouldn't know anything
            # about the pi3hat here, but hey, it's necessary for
            # backward compatibility with the old Pi3HatRouter
            # interface.

            attitude_request = TransportDevice.Request(
                frame=None,
                frame_filter=lambda frame: frame.arbitration_id == -1)

            device_requests[self._pi3hat_device].append(attitude_request)

        if (force_can_check is not None
            and force_can_check != 0
            and self._pi3hat_device):
            # This is another compatibility hack for the pi3hat.
            for bus in range(1, 6):
                if ((1 << bus) & force_can_check) == 0:
                    continue

                force_request = TransportDevice.Request(
                    frame=None,
                    frame_filter=lambda x: True)

                device_requests[self._pi3hat_device].append(force_request)

        # Perform batch transactions in parallel for all devices.
        tasks = []
        device_list = []
        for device, request_list in device_requests.items():
            # We can send at most one broadcast frame that needs
            # replies per transaction.
            broadcast_with_reply = lambda request: (
                (request.frame and ((request.frame.arbitration_id & 0x7f) == 0x7f)) and
                request.frame_filter is not None)

            broadcast_requests = [x for x in request_list
                                  if broadcast_with_reply(x)]
            non_broadcast_requests = [x for x in request_list
                                      if not broadcast_with_reply(x)]

            async def run_in_sequence(d=device,
                                      br=broadcast_requests,
                                      nbr=non_broadcast_requests):

                this_request_attitude = request_attitude
                this_force_can_check = force_can_check

                for b in br:
                    await d.transaction(
                        [b],
                        request_attitude=this_request_attitude,
                        force_can_check=this_force_can_check)

                    this_request_attitude = False
                    this_force_can_check = None

                if nbr:
                    await d.transaction(
                        nbr,
                        request_attitude=this_request_attitude,
                        force_can_check=this_force_can_check)

            if (len(broadcast_requests) + (1 if non_broadcast_requests else 0)) == 1:
                # We can use a single transaction, either because we
                # have a single broadcast request and no other
                # requests, or no broadcast requests.
                tasks.append(device.transaction(
                    request_list,
                    request_attitude=request_attitude,
                    force_can_check=force_can_check))
            else:
                # We have to do multiple transactions to complete
                # this.
                tasks.append(run_in_sequence())

            device_list.append(device)

        # Wait for all transports to complete.
        await asyncio.gather(*tasks)

        # We don't care as much about dropping results in the event of
        # a cancellation here in .cycle.  '.cycle' will mostly be used
        # with register mode commands which are stateless anyways.

        responses = []
        for request in itertools.chain(
                *[request_list for _, request_list in device_requests.items()]):
            for frame in request.responses:
                responses.append(
                    request.command.parse(frame)
                    if (hasattr(request, 'command') and
                        hasattr(request.command, 'parse'))
                    else frame)

        return [self._fix_result_frame(f) for f in responses]

    async def cycle(self, commands,
                    request_attitude=False,
                    read_unsolicited=None,
                    force_can_check=None):
        """Issue all commands, returning any resulting frames.

        Args:

          request_attitude: This is present to allow producing IMU
            data in the same SPI cycle as CAN data with an mjbots
            pi3hat.

          read_unsolicited: An optional list of TransportDevices, for
            which any available unsolicited CAN frames should be
            returned.  If specified on a TransportDevice where moteus
            controllers are commanded, this may result in duplicate
            frame receipts.

          force_can_check: A bitfield to force a connected pi3hat to
            check specific CAN ports.  Modern code should instead use
            the read_unsolicited kwargs.

        """
        results = await self._cycle_batch(
            commands,
            request_attitude=request_attitude,
            force_can_check=force_can_check,
            read_unsolicited=read_unsolicited)
        return results

    async def write(self, command):
        devices = await self._get_devices_for_command(command)
        frame = self._command_to_frame(command)
        if len(devices) == 1:
            await devices[0].send_frame(frame)
        else:
            tasks = [asyncio.create_task(device.send_frame(frame))
                     for device in devices]
            await asyncio.gather(*tasks)

    def _add_task_results_to_cancel_queue(self, tasks):
        to_add = []
        for task in tasks:
            try:
                to_add.append(task.result())
            except:
                pass

        self._add_cancelled_frames(to_add)

    async def read(self, channel=None):
        """Wait for one or more frames to be received.

        This can be used to receive unsolicited data.

        Args:

          channel: If specified, only read from the given
            TransportDevice.
        """
        if self._cancelled_queue:
            return self._fix_result_frame(self._cancelled_queue.pop(0))

        if len(self._devices) == 1:
            return self._fix_result_frame(await self._devices[0].receive_frame())

        if channel is not None:
            assert channel in self._devices
            return self._fix_result_frame(await channel.receive_frame())

        try:
            # Only invoke one receive_frame per parent.
            tasks = [
                asyncio.create_task(device.receive_frame())
                for device in self._parent_devices
            ]
            done, pending = await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED)
        except asyncio.CancelledError:
            # NOTE: We *really* don't want to let frames get lost here
            # in .read().  Unfortunately, since cancellation
            # exceptions can be thrown from nearly everything, it is
            # really hard to ensure that any received frames are not
            # accidentally discarded.  So, we jump through many hoops,
            # including this 'except' block, the 'finally' block and
            # its sub-try/except.

            # Some of our tasks may have still completed.  We should
            # save those so that we can return them later.
            done_tasks = [x for x in tasks if x.done()]
            self._add_task_results_to_cancel_queue(done_tasks)

            raise
        finally:
            # Ensure that all our tasks are either done or cancelled.
            not_done_tasks = [x for x in tasks if not x.done()]
            _ = [x.cancel() for x in not_done_tasks]

            try:
                # We have to wait on those tasks for their
                # cancellation to become final.
                results = await asyncio.shield(
                    asyncio.gather(*not_done_tasks, return_exceptions=True))

                # We are not exiting ourselves, so limit our cancel
                # add to those things that were not done in the main
                # area.
                self._add_cancelled_frames([x for x in results
                                            if isinstance(x, Frame)])
            except asyncio.CancelledError:
                # However, while 'shield' will stop further
                # cancellations from making into our child tasks, we
                # may still be cancelled during this time.  If that
                # happens, we need to try again.
                await asyncio.shield(
                    asyncio.gather(*not_done_tasks, return_exceptions=True))

                # We are definitely going to die, so we need to get
                # *everything* from tasks.
                self._add_task_results_to_cancel_queue(tasks)

                # Now let our own cancellation make it back up.
                raise

        results = [x.result() for x in done]
        if len(results) > 1:
            self._add_cancelled_frames(results[1:])
            del results[1:]

        assert(len(results) == 1)

        return self._fix_result_frame(results[0])

    async def flush_read_queue(self, timeout=0.1, channel=None):
        """Flush any pending receive frames.

        Args:
            timeout: Time in seconds to keep reading before giving up.
            channel: None = all channels
        """

        device_list = self._parent_devices

        if channel is not None:
            assert channel in self._devices
            device_list = [ channel ]

        start_time = asyncio.get_event_loop().time()
        tasks = []
        for device in device_list:
            async def flush_single(device):
                while True:
                    try:
                        remaining = timeout - (asyncio.get_event_loop().time() - start_time)
                        if remaining <= 0:
                            break
                        _ = await asyncio.wait_for(device.receive_frame(),
                                                   timeout=remaining)
                    except asyncio.TimeoutError:
                        break
            tasks.append(flush_single(device))

        if tasks:
            await asyncio.gather(*tasks)

    def _response_to_device_info(self, frame, source):
        # Is this a response to our query?
        source_id = (frame.arbitration_id >> 8) & 0x7f
        dest_id = frame.arbitration_id & 0x7f

        if source_id == 0x7f or dest_id != source:
            # This wasn't addressed to us.
            return None

        # Try to parse and extract the UUID.
        this_device = None
        try:
            result = parse_message(frame)
            uuid_bytes = _extract_uuid_from_result(result)
            if uuid_bytes:
                return DeviceInfo(
                    can_id=source_id,
                    uuid=uuid_bytes,
                    transport_device=frame.channel,
                )
        except RuntimeError:
            pass

        if this_device is None:
            # If parsing failed, we assume this device does
            # not support a UUID.
            return DeviceInfo(
                can_id=source_id,
                transport_device=frame.channel,
            )

        return None

    async def discover(self, can_prefix=0, source=0, timeout=0.2):
        """Discover all controllers attached to any included
        TransportDevices which support discovery.

        Returns:
            A list of DeviceInfo structures containing the CAN ID,
            UUID, and TransportDevice where this device is located.

        """

        # Flush any responses that might be sitting around.
        await self.flush_read_queue()

        # Generate a UUID query.
        query_data, reply_size = _make_uuid_query()

        broadcast_cmd = command.Command()
        broadcast_cmd.destination = 0x7f  # broadcast address
        broadcast_cmd.source = source
        broadcast_cmd.data = query_data
        broadcast_cmd.reply_required = True
        broadcast_cmd.can_prefix = can_prefix
        broadcast_cmd.raw = False
        broadcast_cmd.arbitration_id = None
        broadcast_cmd.expected_reply_size = reply_size

        frame = self._command_to_frame(broadcast_cmd)

        for device in self._devices:
            if not device.empty_bus_tx_safe():
                continue
            await device.send_frame(frame)

        discovered_controllers = []
        start_time = asyncio.get_event_loop().time()

        # Collect from all transport devices in parallel.
        tasks = []
        for device in self._devices:
            if not device.empty_bus_tx_safe():
                continue
            tasks.append(self._read_frames_until_timeout(
                device, start_time + timeout))

        result_lists = await asyncio.gather(*tasks)
        results = []
        for single_list in result_lists:
            results.extend(single_list)

        results = [a for a in
                   [self._response_to_device_info(x, source)
                    for x in results]
                   if a is not None]

        # Now we need go through these and figure out what appropriate
        # DeviceAddress structures are for each.
        for result in results:
            # Is the CAN ID of this device unique within its
            # TransportDevice?  If so, just use that.
            proposed_can_id = result.can_id
            can_conflicts = [
                x for x in results
                if x != result and
                x.transport_device == result.transport_device and
                x.can_id == proposed_can_id]

            if len(can_conflicts) == 0:
                result.address = DeviceAddress(
                    can_id=proposed_can_id,
                    transport_device=result.transport_device)
            elif result.uuid is not None:
                # We have a UUID.  Find the shortest unique prefix.
                for prefix_len in [4, 8, 12, 16]:
                    proposed_prefix = result.uuid[:prefix_len]

                    conflicts = [x for x in results
                                 if x != result and
                                 x.uuid is not None and
                                 x.uuid[:prefix_len] == proposed_prefix]
                    if len(conflicts) == 0:
                        result.address = DeviceAddress(
                            uuid=proposed_prefix,
                            transport_device=result.transport_device)
                        break
            else:
                # We do not have a UUID or a unique CAN-ID.  That
                # means this result gets no address at all as it is
                # not addressable.
                result.address = None

        return sorted(results)



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import glob
import os
import serial
import serial.tools
import serial.tools.list_ports
import sys
import time
import typing

from . import aioserial
from .transport_device import Frame, FrameFilter, TransportDevice


def _hexify(data):
    return ''.join(['{:02X}'.format(x) for x in data])


def _dehexify(data):
    result = b''
    for i in range(0, len(data), 2):
        result += bytes([int(data[i:i + 2], 16)])
    return result


def _find_serial_number(path):
    """Attempt to find the USB serial number for a given device path.

    This function handles both direct device paths and symlinks to the
    actual device, like custom udev rules that create /dev/fdcanusb -> /dev/serial/by-id/foo.
    """
    if not path:
        return None

    try:
        # Resolve symlinks to get the actual device path
        real_path = os.path.realpath(path)

        # Get all ports including symlinks
        ports = serial.tools.list_ports.comports(include_links=True)

        for port in ports:
            # Check if either the device path or the resolved path matches
            if port.device == path or port.device == real_path:
                # Return serial number if available
                if hasattr(port, 'serial_number') and port.serial_number:
                    return port.serial_number

    except Exception:
        # If anything goes wrong, just return None
        pass

    return None


class FdcanusbDevice(TransportDevice):
    """Connects to a single mjbots fdcanusb."""

    def __init__(self, path=None, debug_log=None, disable_brs=False, **kwargs):
        """Constructor.

        Arguments:
          path: serial port where fdcanusb is located
        """
        super(FdcanusbDevice, self).__init__(**kwargs)

        self._disable_brs = disable_brs

        # A fdcanusb ignores the requested baudrate, so we'll just
        # pick something nice and random.
        self._serial = aioserial.AioSerial(port=path, baudrate=9600)

        # Attempt to discover the USB serial number associated with
        # this device for pretty-printing.
        self._serial_number = _find_serial_number(path)

        self._stream_data = b''

        self._ok_waiters = []

        self._reader_task = None
        self._running = False

        self._debug_log = debug_log

        # Start the reader if we can.
        self._start_reader()

    def __repr__(self):
        if self._serial_number:
            return f"Fdcanusb(sn='{self._serial_number}')"
        else:
            return 'Fdcanusb()'

    def close(self):
        if self._reader_task and not self._reader_task.done():
            self._reader_task.cancel()

        if hasattr(self._serial, 'close'):
            self._serial.close()

        self._running = False

    def empty_bus_tx_safe(self):
        return True

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def _start_reader(self):
        self._running = True
        try:
            self._reader_task = asyncio.create_task(self._reader_loop())
        except RuntimeError:
            self._reader_task = None

    async def _ensure_reader_started(self):
        if self._reader_task is None and self._running:
            self._reader_task = asyncio.create_task(self._reader_loop())

    async def _reader_loop(self):
        try:
            while self._running:
                try:
                    line = await self._readline(self._serial)
                    if not line:
                        continue

                    if self._debug_log:
                        self._write_log(b'< ' + line.rstrip())

                    if line.startswith(b'rcv'):
                        frame = self._parse_frame(line)
                        await self._handle_received_frame(frame)
                    elif line.startswith(b'OK'):
                        await self._handle_ok_response(line)
                    else:
                        await self._handle_other_response(line)
                except asyncio.CancelledError:
                    break
                except Exception as e:
                    raise
                    # Log and continue running.
                    if self._debug_log:
                        self._write_log(f'ERROR: {str(e)}'.encode('latin1'))

                    # Sleep briefly to prevent a tight error loop.
                    await asyncio.sleep(0.01)
        finally:
            self._running = False

    async def _handle_ok_response(self, line):
        # Just notify the first non-done OK waiter.  An OK received
        # with no waiters is assumed to be stale.
        #
        # Callers are *required* to enqueue their self._ok_waiters
        # *before* sending anything that could result in an OK being
        # emitted.
        for waiter in self._ok_waiters:
            if not waiter.done():
                waiter.set_result(None)
                return

    async def _handle_other_response(self, line):
        raise RuntimeError(f'{self} received error {line}')

    async def _readline(self, stream):
        while True:
            offset = min((self._stream_data.find(c) for c in b"\r\n"
                          if c in self._stream_data), default=None)
            if offset is not None:
                to_return, self._stream_data = (
                    self._stream_data[0:offset+1],
                    self._stream_data[offset+1:])
                if offset > 0:
                    return to_return
                else:
                    continue
            else:
                data = await stream.read(8192, block=False)
                if not data:
                    # If for some reason we got a completely empty
                    # response, which shouldn't happen, ensure we
                    # yield control to somebody else so that hopefully
                    # we eventually make forward progress.
                    await asyncio.sleep(0)
                else:
                    self._stream_data += data

    def _parse_frame(self, line):
        fields = line.split(b" ")
        frame = Frame()
        frame.data = _dehexify(fields[2])
        frame.dlc = len(frame.data)
        frame.arbitration_id = int(fields[1], 16)
        frame.channel = self

        for flag in fields[3:]:
            if flag == b'E':
                frame.is_extended_id = True
            if flag == b'B':
                frame.bitrate_switch = True
            if flag == b'F':
                frame.is_fd = True

        return frame

    async def _write_send_frame(self, frame):
        actual_brs = frame.bitrate_switch and not self._disable_brs

        hexdata = _hexify(frame.data)
        on_wire_size = self._round_up_dlc(len(frame.data))
        hexdata += self._padding_hex * (on_wire_size - len(frame.data))

        flags = ''
        if actual_brs:
            flags += 'B'
        else:
            flags += 'b'
        if frame.is_fd:
            flags += 'F'

        cmd = "can send {:04x} {}{}{}\n".format(
            frame.arbitration_id,
            hexdata,
            ' ' if len(flags) > 0 else '',
            flags).encode('latin1')

        self._serial.write(cmd)
        if self._debug_log:
            self._write_log(b'> ' + cmd.rstrip())

    async def send_frame(self, frame: Frame):
        await self._ensure_reader_started()

        try:
            ok_waiter = asyncio.Future()
            self._ok_waiters.append(ok_waiter)

            await self._write_send_frame(frame)
            await self._serial.drain()

            await ok_waiter
        finally:
            self._ok_waiters = [w for w in self._ok_waiters
                                if w != ok_waiter]

    async def receive_frame(self):
        await self._ensure_reader_started()

        return await super(FdcanusbDevice, self).receive_frame()

    async def transaction(
            self,
            requests: typing.List[TransportDevice.Request],
            **kwargs):

        # We do not support child devices.
        assert not any([request.child_device is not None
                        for request in requests])

        def make_subscription(request):
            future = asyncio.Future()

            async def handler(frame, request=request, future=future):
                if future.done():
                    # Stick it in our receive queue so that it isn't
                    # lost.
                    self._receive_queue.append(frame)

                    return

                request.responses.append(frame)
                # While we may receive more than one frame for a given
                # request, we only wait for one.
                future.set_result(None)

            return self._subscribe(request.frame_filter, handler), future

        subscriptions = [
            make_subscription(request)
            for request in requests
            if request.frame_filter is not None
        ]

        try:
            # Now we will send all our requests.
            ok_waiters = set([asyncio.Future()
                              for _ in range(len(requests))])
            try:
                self._ok_waiters.extend(list(ok_waiters))

                # TODO: Provide control over the amount of pipelining
                # like the C++ class.
                for request in requests:
                    await self._write_send_frame(request.frame)

                await self._serial.drain()

                await asyncio.gather(*ok_waiters)
            finally:
                self._ok_waiters = [w for w in self._ok_waiters
                                    if w not in ok_waiters]

            # If there any responses to wait for, do so.
            if subscriptions:
                await asyncio.gather(*[x[1] for x in subscriptions])
        finally:
            # Clean up all our subscriptions.
            for x in subscriptions:
                x[0].cancel()

    def _write_log(self, output: bytes):
        assert self._debug_log is not None
        self._debug_log.write(f'{time.time():.6f}/{self._serial_number} '.encode('latin1') + output + b'\n')

    @staticmethod
    def detect_fdcanusbs():
        if sys.platform == 'win32':
            return FdcanusbDevice.pyserial_detect_fdcanusbs()

        maybe_list = glob.glob('/dev/serial/by-id/*fdcanusb*')
        if len(maybe_list):
            return sorted(maybe_list)

        return FdcanusbDevice.pyserial_detect_fdcanusbs()

    @staticmethod
    def detect_fdcanusb():
        return FdcanusbDevice.detect_fdcanusbs()[0]

    @staticmethod
    def pyserial_detect_fdcanusbs():
        ports = serial.tools.list_ports.comports()

        return [x.device for x in ports
                if x.vid == 0x0483 and x.pid == 0x5740]



# Copyright 2025 mjbots Robotic Systems, LLC.  info@mjbots.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import collections
import os
import sys
import time
import typing

from .transport_device import Frame, FrameFilter, TransportDevice, Subscription

can = None


class PythonCanSubscription(Subscription):
    def __init__(self, device, subscription_id):
        self._device = device
        self._id = subscription_id

    def cancel(self):
        self._device._remove_subscription(self._id)


class PythonCanDevice(TransportDevice):
    '''Implements a 'Transport' on top of python-can.'''

    def __init__(self, *args, **kwargs):
        '''Nearly all arguments pass through to can.Bus'''

        super(PythonCanDevice, self).__init__(
            **{k: kwargs[k] for k in ['max_buffer_size',
                                      'padding_hex'] if k in kwargs})

        self._padding_byte = bytes.fromhex(self._padding_hex)
        self._debug_log = kwargs.pop('debug_log', None)

        # We delay importing this until we need it, as it can take a
        # while.
        global can
        if not can:
            import can
            try:
                can.rc = can.util.load_config()
            except can.CanInterfaceNotImplementedError as e:
                if 'Unknown interface type "None"' not in str(e):
                    raise

        if 'fd' not in can.rc:
            can.rc['fd'] = True

        self._disable_brs = False
        if 'disable_brs' in kwargs:
            self._disable_brs = kwargs['disable_brs']
            del kwargs['disable_brs']

        if ('timing' not in kwargs and
            kwargs.get('interface', can.rc.get('interface', None)) == 'pcan'):
            # We default to the timing that works with moteus and assume
            # an 80MHz base clock, which seems pretty typical for PCAN
            # interfaces.
            kwargs['timing'] = can.BitTimingFd.from_sample_point(
                nom_bitrate=1000000, data_bitrate=5000000,
                nom_sample_point=66, data_sample_point=66,
                f_clock=80000000)

        self._can = can.Bus(*args, **kwargs)
        self._setup = False

        self._notifier = None

        self._log_prefix = (
            getattr(self._can, 'channel', None) or
            kwargs.get('channel', None) or
            str(self._can))

    def __repr__(self):
        return f"PythonCan('{self._log_prefix}')"

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        if self._notifier:
            self._notifier.stop()
            self._notifier = None

        self._can.shutdown()

    def empty_bus_tx_safe(self):
        return True

    def _maybe_setup(self):
        if self._setup:
            return

        self._notifier = can.Notifier(
            self._can, [self._receive_handler],
            loop=asyncio.get_event_loop())
        self._setup = True

    async def _receive_handler(self, message):
        if message.is_error_frame:
            return

        frame = self._can_message_to_frame(message)

        if self._debug_log:
            self._write_log(f'< {frame.arbitration_id:04X} {frame.data.hex().upper()}'.encode('latin1'))

        await self._handle_received_frame(frame)

    def _frame_to_can_message(self, frame: Frame):
        dlc = self._round_up_dlc(len(frame.data))
        padding_bytes = dlc - len(frame.data)

        return can.Message(
            arbitration_id=frame.arbitration_id,
            is_extended_id=frame.is_extended_id,
            dlc=dlc,
            data=frame.data + bytes(self._padding_byte) * padding_bytes,
            is_fd=frame.is_fd,
            bitrate_switch=frame.bitrate_switch and not self._disable_brs,
        )

    def _can_message_to_frame(self, message) -> Frame:
        return Frame(
            arbitration_id=message.arbitration_id,
            data=message.data,
            dlc=message.dlc,
            is_extended_id=message.is_extended_id,
            is_fd=message.is_fd,
            bitrate_switch=getattr(message, 'bitrate_switch', False),
            channel=self,
        )

    async def send_frame(self, frame: Frame):
        self._maybe_setup()

        can_message = self._frame_to_can_message(frame)

        if self._debug_log:
            self._write_log(f'> {frame.arbitration_id:04x} {can_message.data.hex().upper()}'.encode('latin1'))

        self._can.send(self._frame_to_can_message(frame))

    async def receive_frame(self):
        self._maybe_setup()

        return await super(PythonCanDevice, self).receive_frame()

    async def transaction(
            self,
            requests: typing.List[TransportDevice.Request],
            **kwargs):
        # We do not support child devices.
        assert not any([request.child_device is not None
                        for request in requests])

        self._maybe_setup()

        def make_subscription(request):
            future = asyncio.Future()

            async def handler(frame, request=request, future=future):
                if future.done():
                    return

                request.responses.append(frame)
                future.set_result(None)

            return self._subscribe(request.frame_filter, handler), future

        subscriptions = [
            make_subscription(request)
            for request in requests
            if request.frame_filter is not None
        ]

        try:
            for request in requests:
                await self.send_frame(request.frame)

            if subscriptions:
                await asyncio.gather(*[x[1] for x in subscriptions])
        finally:
            for x in subscriptions:
                x[0].cancel()

    def _write_log(self, output: bytes):
        assert self._debug_log is not None
        self._debug_log.write(f'{time.time():.6f}/{self._log_prefix} '.encode('latin1') + output + b'\n')

    @staticmethod
    def enumerate_devices(**kwargs):
        global can
        if not can:
            import can

        import logging

        # python-can spews a lot of logging during enumeration.
        # Squelch as much of that as we can.

        log = logging.getLogger("can")  # covers can.interface and sub-loggers
        prev_level, prev_prop = log.level, log.propagate
        log.setLevel(logging.CRITICAL)

        # Don't bubble to root handlers.
        log.propagate = False

        try:
            # Try to auto-enumerate.
            interfaces_to_check = os.getenv(
                "MJBOTS_PYTHONCAN", "socketcan,pcan,kvaser,vector").split(',')
            available_configs = can.detect_available_configs(interfaces_to_check)

            if not available_configs:
                raise RuntimeError("No python-can interfaces detected")

            def _merge_args(a, b):
                return {**a, **b}

            return [PythonCanDevice(**_merge_args(kwargs, x))
                    for x in available_configs]
        finally:
            # Restore logging to the previous state.
            log.setLevel(prev_level)
            log.propagate = prev_prop
